/* bb2ns.c
   vim:ts=4
   Daemon to accept Big Brother logs, parse them, and send them to 
   Nagios via NSCA.
 
   Call from (x)inetd!

   Problems: validating host/service names.  Mapping host names.  Multithread.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <unistd.h>
#include <netinet/in.h>
#include <time.h>
#include <fcntl.h>
#include <dirent.h>

#define BBPORT 1984
#define NAGCMD "/u01/nagios/log/rw/nagios.cmd"
#define FD_MAX 64
#define ALERTSERVICE "Big Brother gateway"
#define BBDIR "/u01/www/html/bbgw"
#undef BBDIR

extern char *optarg;
extern int optind, opterr, optopt;

char *cmdfile, *logfile;
int port;
int debug = 0;
char defcmdfile[] = NAGCMD;
char debugcmdfile[] = "/tmp/nagios.cmd";
int sock;
int inetd;
char defalertservice[] = ALERTSERVICE;
char *alerthost, *alertservice;

/* ----------------------------------------------------------------- */

#ifdef BBDIR
void writehtml(char *bbhost, char *bbsvc, char *page) {
	char filename[256];
	char ifilename[256];
	FILE *hfp, *ifp;
	DIR *dirp;
	struct dirent *sdp;
	char *p,*q, *r;
	char newmsg[4096];

	sprintf(filename,"%s/%s",BBDIR,bbhost);
	dirp = opendir(filename);
	if( !dirp ) { 
		mkdir(filename,0755); 
		/* recreate top index */
		sprintf(ifilename,"%s/index.html",BBDIR);
		ifp = fopen(ifilename,"w"); if(!ifp) { return; }
		dirp = opendir(BBDIR); if(!dirp) { fclose(ifp); return; }
		fprintf(ifp,"<HTML><HEAD><TITLE>BB Status</TITLE></HEAD>\n<BODY>\n");
		fprintf(ifp,"<h1>Big Brother hosts</h1>\n<UL>");
		while( sdp = readdir( dirp ) ) {
			if( !strpbrk( sdp->d_name,"." ) ) {
				fprintf(ifp,"<LI><A href=%s/index.html>%s</A></LI>\n",
					sdp->d_name,sdp->d_name);
			}
		}
		fprintf(ifp,"</UL></BODY></HTML>\n");
		fclose(ifp);
		closedir(dirp);
		dirp = opendir(filename); 
	}
	sprintf(filename,"%s/%s/%s.html",BBDIR,bbhost,bbsvc);
	hfp = fopen(filename,"r"); 
	if(!hfp) { 
		hfp = fopen(filename,"w"); if(!hfp) { return; }
		/* recreate host index */
		sprintf(ifilename,"%s/%s/index.html",BBDIR,bbhost);
		ifp = fopen(ifilename,"w"); if(!ifp) { closedir(dirp); return; }
		fprintf(ifp,"<HTML><HEAD><TITLE>%s</TITLE></HEAD>\n<BODY>\n",bbhost);
		fprintf(ifp,"<h1>%s Status</h1>\n<UL>",bbhost);
		while( sdp = readdir( dirp ) ) {
			if( ( p = strstr(sdp->d_name,".html") )
				&& strcmp(p,"index") ) {
				*p = (char)0;
				fprintf(ifp,"<LI><A href=%s.html>%s</A></LI>\n",
					sdp->d_name,sdp->d_name);
			}
		}
		fprintf(hfp,"</UL><HR><A href=../index.html>Other hosts</A><HR>\n");
		fprintf(ifp,"</BODY></HTML>\n");
		fclose(ifp);
	} else { 
		fclose(hfp); 
		hfp = fopen(filename,"w"); if(!hfp) { return; }
	}
	closedir(dirp); /* dont need it any more */

	/* clean up the message */
	q = newmsg;
	*q = (char)0;
	while( p = strpbrk(r,"&") ) {
		strncpy(q,r,(p-r)); q+=(p-r); *q = (char)0; r = p;
		if(!strncmp(r,"&red",4)) {
			strcat(q,"<IMG src=../red.gif    height=16 width=16 border=0>");
			q+=50; r+=4;
		} else if( !strncmp(r,"&yellow",7) ) {
			strcat(q,"<IMG src=../yellow.gif height=16 width=16 border=0>");
			q+=50; r+=7;
		} else if( !strncmp(r,"&green",6) ) {
			strcat(q,"<IMG src=../green.gif  height=16 width=16 border=0>");
			q+=50; r+=6;
		} else {
			r+=1; *q++ = '&'; *q = (char)0;
		}
	}
	strcat(q,r);

	/* print it out */
	fprintf(hfp,"<HTML><HEAD><TITLE>%s: %s</TITLE></HEAD>\n<BODY>\n",
		bbhost,bbsvc);
	fprintf(hfp,"<H1>%s</H1>\n<H2>%s</H2>\n<HR><PRE>%s</PRE><HR>\n",
		bbhost,bbsvc,page);

	fprintf(hfp,"<A href=index.html>Other data for this host</A><BR>\n");
	fprintf(hfp,"<A href=../index.html>Other hosts</A><HR>\n");
	fprintf(hfp,"</BODY></HTML>\n");
	fclose( hfp ); 
}
#endif

void logmsg(char *msg) {
	FILE *lfp;
	struct tm *stmp;
	time_t t;

	if(!logfile) { return; }
	/* write msg to the logfile, with a timestamp */
	lfp = fopen(logfile,"a");
	if(!lfp) { return; }
	t = time(NULL);
	stmp = localtime(&t);
	fprintf(lfp,"%02d:%02d\t%s\n",stmp->tm_hour,stmp->tm_min,msg);
	fclose(lfp);
/*	if(alerthost) { sendalert(time(NULL),alerthost,alertservice,1,msg); } */
}
void logmsg2(char *msg,char *arg) {
	FILE *lfp;
	struct tm *stmp;
	time_t t;
	char buf[256];

	if(!logfile) { return; }
	/* write msg to the logfile, with a timestamp */
	lfp = fopen(logfile,"a");
	if(!lfp) { return; }
	t = time(NULL);
	stmp = localtime(&t);
	fprintf(lfp,"%02d:%02d\t",stmp->tm_hour,stmp->tm_min);
	sprintf(buf,msg,arg);
	fprintf(lfp,"%s\n",buf);
	fclose(lfp);
/*	if(alerthost) { sendalert(time(NULL),alerthost,alertservice,1,buf); }*/
}

void readconfig(char *f) {
	/* here, we parse the Nagios config, if necessary */
}
int verifysvc(char *h, char *svc) {
	return 1;
}

void dohelp() {
	printf( "Usage: bb2ns [-i][-d level][-p port][-l logfile][-c nagios.cfg][-f commandfile]\n             [-H alerthost [-S alertservice]]\n");
	printf( "Debug levels: -d 0=errors(default), 1=basic, 2=detail, 3=data\n");
	printf( "Alert host is what host/service to generate a Nagios alert on if there is an\nerror in the processing.\n");
	exit(0);
}

int sendalert(int ts, char *h, char *svc, int stat, char *msg) {
	int cmdfd;
	int rv;
	char buf[2048];

	if(debug>1) {
	 	logmsg2("Sending [%s]",msg); 
	}
	/*	cmdfd = open(cmdfile,O_WRONLY|O_CREAT|O_APPEND,S_IRWXU); */
	cmdfd = open(cmdfile,O_WRONLY);
	if(cmdfd<0) {
		logmsg2("Unable to open command file %s for writing",cmdfile);
		return;
	}
	if(msg[strlen(msg)-1]=='\n') {
		msg[strlen(msg)-1] = (char)0; /* lose the last character */
	}
	sprintf(buf,"[%d] PROCESS_SERVICE_CHECK_RESULT;%s;%s;%d;%s\n",
		ts,h,svc,stat,msg);
/*	if(debug) { flock( cmdfd, LOCK_EX );} */
	write(cmdfd,buf,strlen(buf));
/*	if(debug) { flock( cmdfd, LOCK_UN );} */
	close(cmdfd);
	if(debug>1) { logmsg2("Sent [%s]",buf); }
}

void rmvpbr(char *s) {
	for( ;*s;s++ ) {
		if(! strncmp(s, "<BR>",4 ) ) { memcpy(s,"   \n",4); continue; }
		if(! strncmp(s, "<br>",4 ) ) { memcpy(s,"   \n",4); continue; }
		if(! strncmp(s, "<P>",3 ) ) { memcpy(s,"  \n",3); continue; }
		if(! strncmp(s, "<p>",3 ) ) { memcpy(s,"  \n",3); continue; }
		if(! strncmp(s, "<TABLE>",7 ) ) { memcpy(s,"      \n",7); continue; }
		if(! strncmp(s, "</TABLE>",8 ) ) { memcpy(s,"       \n",8); continue; }
	}
}

/* clean the string, make it HTML safe */
void cleanup(char *s, int len) {
	char *x,*s2,*t, *y;
	char tmp[2048];
	int l;

	s2 = s; t = tmp; tmp[0] = (char)0;
	/* remove < and > replace \n with <BR> */
	/* replace &(red|yellow|green) with [&1] */
	while( x = strpbrk(s2,"&<>\n") ) {
		l = x-s2;
		strncpy(t,s2,l); t+=l; s2=x+1;
		if( *x == '<' ) { /* strip HTML */
			if( y = strpbrk(s2,">") ) {
				if( (*(x+1) == 'T') && (*(x+2)=='D') ) {*(t++)=' ';}
				s2 = y + 1;
			} else {
				strncpy(t,"&lt;",4); t+=4;
			}
		} else if( *x == '>' ) {
			strncpy(t,"&gt;",4); t+=4;
		} else if( *x == '\n' ) {
			strncpy(t,"<BR>",4); t+=4;
		} else if( *x == '&' ) {
			if(! strncmp(s2,"red",3) ) {
				strncpy(t,"[red]",5); t+=5; s2+=3;
			} else if(! strncmp(s2,"yellow",6) ) {
				strncpy(t,"[yellow]",8); t+=8; s2+=6;
			} else if(! strncmp(s2,"green",5) ) {
				strncpy(t,"[green]",7); t+=7; s2+=5;
			} else {
				strncpy(t,"&amp;",5); t+=5;
			}
		}
	}
	*t = (char) 0;
	strcat(tmp,s2); /* the bit left */
	strncpy(s,tmp,len);
	s[len-1] = (char)0;
	len = strlen(s);
	while((len>3) && !strncmp(s+len-4,"<BR>",4) ) {
		len -= 4; s[len] = (char)0; 
	}
}

void selectdisk(char *s) {
	char tmp[4096]; char *t; char *fs, *pc;
	char *tok; int fld;
	tmp[0] = (char)0;
	for(t=(char *)strtok(s,"\n");t;t=(char *)strtok((char *)NULL,"\n")) {
		if(debug>2) { logmsg2("parse[%s]",t); } 
		if(strstr(t,"/dev")) {  /* UNIX */
			fs = pc = NULL; fld = 0;
			for( tok=t; *tok; ) {
				if( (*tok != ' ')&&(*tok != '\t') ) {
					if( fld == 5 ) { fs = tok; }
					if( fld == 4 ) { pc = tok; }
					while( *tok && ((*tok!=' ')&&(*tok!='\t'))) { tok++; } 
					if(*tok){*(tok++) = '\0';} /* terminate token */
					if( fld == 5 ) { break; }
					fld++;
				}
				while( (*tok == ' ') || (*tok == '\t') ) 
					{ *(tok++)='\0'; } /*skip trailing space */
			}
			if(fs && pc) {
			strcat(tmp,fs); strcat(tmp,": "); strcat(tmp,pc); strcat(tmp," ");
			}
			continue;
		}
		if(strstr(t,"/FIXED/")) {  /* windows */
			fs = pc = NULL; fld = 0;
			for( tok=t; *tok; ) {
				if( (*tok != ' ')&&(*tok != '\t') ) {
					if( fld == 0 ) { fs = tok; }
					if( fld == 4 ) { pc = tok; }
					while( *tok && ((*tok!=' ')&&(*tok!='\t'))) { tok++; } 
					if(*tok){*(tok++) = '\0';} /* terminate token */
					if( fld == 4 ) { break; }
					fld++;
				}
				while( (*tok == ' ') || (*tok == '\t') ) 
					{ *(tok++)='\0'; } /*skip trailing space */
			}
			if(fs && pc) {
			strcat(tmp,fs); strcat(tmp,": "); strcat(tmp,pc); strcat(tmp," ");
			}
			continue;
		}
		if(strstr(t,"Inodes:")) { break; }
	}
	if(!*tmp) { strcpy(tmp,"No filesystems found"); }
	strcpy(s,tmp);
}
void selectprocs(char *s) {
	char tmp[4096]; char *t;
	tmp[0] = (char)0;
	for(t=(char *)strtok(s,"\n");t;t=(char *)strtok((char *)NULL,"\n")) {
		if(strstr(t,"DOWN")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
	}
	if(!*tmp) { strcpy(tmp,"All processes OK"); }
	strcpy(s,tmp);
}
void selectcpu(char *s) {
	char tmp[4096]; char *t;
	tmp[0] = (char)0;
	for(t=(char *)strtok(s,"\n");t;t=(char *)strtok((char *)NULL,"\n")) {
		if(strstr(t,"CPU states")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
		if(strstr(t,"up:")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
	}
	if(!tmp[0]) { strcpy(tmp,"CPU status from Big Brother"); }
	strcpy(s,tmp);
}
void selectlines(char *s,int first,int green,int yellow,int red) {
	char tmp[4096];
	char *t;
	tmp[0] = (char)0;
	for(t=(char *)strtok(s,"\n");t;t=(char *)strtok((char *)NULL,"\n")) {
		if(first) { first = 0;
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
		if(green && strstr(t,"&green")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
		if(yellow && strstr(t,"&yellow")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
		if(red && strstr(t,"&red")) { 
			strcat(tmp,t); strcat(tmp,"\n"); continue; }
	}
	if(!tmp[0]) { strcpy(tmp,"All OK"); }
	strcpy(s,tmp);
}

/* process request on fd rp/wp.  Return TRUE if finished */
int dorequest(int rp, int wp) {
	char buffer[4096];
	int ts,rv,nc;
	char host[64], service[64], colour[32],has[128];
	int status,offset;
	char msg[1024], *msgp;
	char *sp, *dstart;
	struct tm stm;

	/* first, read in the command, and anything that goes with it */
	rv = read(rp, buffer, sizeof(buffer) );
	if( rv < 0 ) {
		logmsg("Error on read");
		return 1;
	}
	if( !rv ) { return 1; }

	if(msg[strlen(buffer)-1]=='\n') {
		msg[strlen(buffer)-1] = (char)0; /* lose the last character */
	}

	if(debug>0) { 
		logmsg("Incoming message received");
		if(debug>2) { logmsg(buffer);   }
		if(!inetd) { printf("Recv: %s\n",buffer); }  
	}

	/* now, check that it is a valid command, and split up the data */
	if(strncmp(buffer,"status ",7)) { 
		if(debug>0) { logmsg("Not a status message..."); }
		return 1; 
	}

	has[0] = host[0] = service[0] = colour[0] = (char)0; status = -1;
	offset = 7;
	sscanf(buffer+offset,"%s %s %n",has,colour,&nc);
	offset+=nc;
	if(*has) { strncpy(host,(char *)strtok(has,"."),sizeof(host)); }
	if(*has && *host) { 
		strncpy(service,(char *)strtok(NULL," "),sizeof(service)); }

	if(!strcmp(colour,"red")) { status = 2; } 
	else if(!strcmp(colour,"green")) { status = 0; } 
	else if(!strcmp(colour,"yellow")) { status = 1; } 
	else if(!strcmp(colour,"purple")) { status = 3; }
	if(!(dstart=(char *)strptime(buffer+offset, "%c", &stm))) {
	if(!(dstart=(char *)strptime(buffer+offset, "%Ec", &stm))) {
	  if(!(dstart=(char *)strptime(buffer+offset, "%a %b%n%d %H:%M:%S%n%Y", &stm))) {
		if(!(dstart=(char *)strptime(buffer+offset, "%a %b%n%d %H:%M:%S%nNZDT%n%Y", &stm))){
			dstart=(char *)strptime(buffer+offset, "%a %b%n%d %H:%M:%S%nNZST%n%Y", &stm);
		}
	  }
	}
	}
	if(!dstart) { dstart = buffer + offset + 24; }
	ts = mktime(&stm);
	strncpy(msg,dstart,sizeof(msg)-1);
	msg[sizeof(msg)-1] = (char)0; /* just in case */
	
	/* sanity check */
	if( !*host    ) { logmsg("Invalid host"     ); return 1; }
	if( !*service ) { logmsg("Invalid service"  ); return 1; }
	if( status<0  ) { logmsg2("Invalid status [%s]",colour); return 1; }
	if( ts < 1    ) { logmsg2("Invalid timestamp [%.25s]",buffer+offset); return 1; }
#ifdef BBDIR
	writehtml(host,service,msg);
#endif
	/* check if it is a valid service */
	if( ! verifysvc(host,service) ) { 
		if(debug>0) { logmsg("Not a recognised service name"); }
		if(alerthost) {
			sendalert(time(NULL),alerthost,alertservice,1,
				"Unrecognised service name");
		}
		return 1; 
	}

	msgp = msg;
	while( *msgp && (*msgp==' ')) { msgp++; } /* skip leading spaces */
	if(*msgp == '[') { 
		msgp = strchr(msgp,']'); 
		if(msgp) { msgp++; } else { msgp = msg; }
	} /* skip windows hostname */
	if(msg != msgp ) { strcpy(msg,msgp); msgp=msg; }

	rmvpbr(msgp); /* clean up special HTML */

	/* if it is something special, the do appropriate parsing of output */
	if(!strcmp(service,"cpu")) {
		if(debug>2) { logmsg2("Selecting [%s]",msgp);  }
		selectcpu(msgp);
	} else if(!strcmp(service,"procs")) {
		selectprocs(msgp);
	} else if(!strcmp(service,"disk")) {
		if(debug>2) { logmsg2("Selecting [%s]",msgp); }
		selectdisk(msgp);
	} else if(!strcmp(service,"msgs")) {
/*		selectlines(msg,1,0,0,0); */
	} else if(!strcmp(service,"tux")) {
		selectlines(msgp,0,0,1,1);
	} else if(!strcmp(service,"oracle2")) {
		selectlines(msgp,1,1,1,1);
	} else if(!strcmp(service,"mem")) {
		if(debug>2) { logmsg2("Selecting [%s]",msgp);  }
		selectlines(msgp,1,1,1,1);
	} else if(!strcmp(service,"queue")) {
		selectlines(msgp,1,1,1,1);
	}
/*	if(debug) { logmsg2("Cleaning [%s]",msg); } */
	cleanup(msgp,sizeof(msg));
	if(debug>2) {	logmsg2("Cleaned [%s]",msgp);  }

	/* now send the alert */
	sendalert(ts,host,service,status,msgp);

	/* and exit */
	return 1;
}

void doserver() {
	fd_set rfds, wfds, xfds;
	fd_set rfd, wfd, xfd;
	int i,rv;
	struct sockaddr ssa;
	socklen_t	sl;

	FD_ZERO(&rfds); FD_ZERO(&wfds); FD_ZERO(&xfds);

	FD_SET(sock,&rfds); FD_SET(sock,&xfds);
	
	for(;;) {
		rfd = rfds; wfd = wfds; xfd = xfds;
		rv = select(FD_MAX , &rfd, &wfd, &xfd, (struct timeval *)0 );
		if(!rv) { break; }
		if( FD_ISSET(sock,&rfd) ) {
			/* new connection */
			i = accept(sock,&ssa,&sl);
			FD_SET(i,&rfds);
		}
		for(i=0;i<FD_MAX;i++) {
			if(FD_ISSET(i,&rfd)) {
				if( dorequest(i,i) ) {
					close(i);
					FD_CLR(i,&rfds);
				}
			}
		}
	}
}

/* ----------------------------------------------------------------- */

int main(int argc, char *argv[]) {
	int c;
	struct sockaddr_in ssi;
	int val;
	char *config;

	/* process any arguments */
	config = (char *)0; logfile = (char *)0; inetd = 0;
	alerthost = (char *)0; alertservice = defalertservice;
	cmdfile = defcmdfile;
	port = BBPORT;
	while((c = getopt(argc,argv,"hd:p:c:f:l:iH:S:")) != -1) {
		switch(c) {
			case 'd': debug=atoi(optarg); break;
			case 'p': port=atoi(optarg); break;
			case 'c': config=optarg; break;
			case 'f': cmdfile=optarg; break;
			case 'l': logfile=optarg; break;
			case 'i': inetd=1; break;
			case 'H': alerthost=optarg; break;
			case 'S': alertservice=optarg; break;
			case 'h': 
			case '?': dohelp(); break;
		}
	}

	if(config) { readconfig(config); }

/*	if(debug) { cmdfile = debugcmdfile; } */

	/* open BB socket to listen */
	if(!inetd) {
		logmsg("Opening logfile");
		sock = socket(PF_INET,SOCK_STREAM,0);
		if( sock < 0 ) { perror("socket"); exit(1); }
		memset(&ssi,0,sizeof(ssi));
		ssi.sin_family = AF_INET; 
		ssi.sin_port   = port;
		ssi.sin_addr.s_addr = INADDR_ANY; /* bind to all addresses */
		if( bind( sock, (struct sockaddr *)&ssi, sizeof(ssi) ) ) {
			perror("bind"); exit(1);
		}
		listen( sock, 20 ); /* set queue length */
		/* val = 1; setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE,&val,sizeof(val)); */
		
		/* background and disassociate */
		fclose(stdin); 
/*		if(!debug) { */
			if(fork()) { printf("BB2NS daemon starting...\n"); exit(0); }
			fclose(stdout); fclose(stderr);
/*		} else {
			printf("BB2NS daemon starting in foreground\n");
			val = 1; setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&val,sizeof(val));
		} */
		logmsg("bb2ns daemon starting");
		if(alerthost) {
			sendalert(time(NULL),alerthost,alertservice,0,
				"Gateway starting");
		}
	}
	
	/* process */
	if(inetd) {
		dorequest(fileno(stdin),fileno(stdout));
	} else {
		doserver;
		if(debug) { printf("Exiting...\n"); }
		if(alerthost) {
			sendalert(time(NULL),alerthost,alertservice,2,"Gateway shut down");
		}
	}

	exit(0);
}
